import os
try:
    import ssl
except ImportError:
    ssl = None
import socket
import unittest
from unittest import mock

import pytest

from blackhole.config import Config
from blackhole.control import (_context, _socket, server, setuid, setgid,
                               pid_permissions)
from blackhole.exceptions import BlackholeRuntimeException

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_tls_context_no_config():
    ctx = _context()
    assert ctx is None


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_tls_context_no_dhparams():
    tls_cert = create_file('cert.cert')
    tls_key = create_file('key.key')
    cfile = create_config(('listen=127.0.0.1:25', 'tls_listen=127.0.0.1:9000',
                           'tls_cert={}'.format(tls_cert),
                           'tls_key={}'.format(tls_key),))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('ssl.SSLContext.load_cert_chain'):
        _context(use_tls=True)


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_tls_context_less_secure():
    tls_cert = create_file('cert.cert')
    tls_key = create_file('key.key')
    cfile = create_config(('listen=127.0.0.1:25', 'tls_listen=127.0.0.1:9000',
                           'tls_cert={}'.format(tls_cert),
                           'tls_key={}'.format(tls_key),))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', True), ))
    with mock.patch('ssl.SSLContext.load_cert_chain'):
        _context(use_tls=True)


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_tls_context_dhparams():
    tls_cert = create_file('cert.cert')
    tls_key = create_file('key.key')
    tls_dhparams = create_file('dhparams.pem')
    cfile = create_config(('listen=127.0.0.1:25', 'tls_listen=127.0.0.1:9000',
                           'tls_cert={}'.format(tls_cert),
                           'tls_key={}'.format(tls_key),
                           'tls_dhparams={}'.format(tls_dhparams)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('ssl.SSLContext.load_cert_chain'), \
            mock.patch('ssl.SSLContext.load_dh_params') as dh:
        _context(use_tls=True)
    assert dh.called is True


@unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_ipv6_socket_fails():
    with mock.patch('socket.socket.bind', side_effect=OSError), \
            pytest.raises(BlackholeRuntimeException):
        _socket('::', 25, socket.AF_INET)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_ipv4_socket_fails():
    with mock.patch('socket.socket.bind', side_effect=OSError), \
            pytest.raises(BlackholeRuntimeException):
        _socket('127.0.0.1', 25, socket.AF_INET)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_ipv4_bind_fails():
    cfile = create_config(('listen=127.0.0.1:9000',))
    Config(cfile).load()
    with mock.patch('socket.socket.bind',
                    side_effect=OSError) as mock_sock, \
            pytest.raises(BlackholeRuntimeException):
        server('127.0.0.1', 9000, socket.AF_INET, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_ipv6_bind_fails():
    cfile = create_config(('listen=:::9000',))
    Config(cfile).load()
    with mock.patch('socket.socket.bind',
                    side_effect=OSError) as mock_sock, \
            pytest.raises(BlackholeRuntimeException):
        server('::', 9000, socket.AF_INET6, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@mock.patch('socket.socket.bind')
def test_create_server_ipv4_bind_works(mock_sock):
    cfile = create_config(('listen=127.0.0.1:9000',))
    Config(cfile).load()
    server('127.0.0.1', 9000, socket.AF_INET, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_ipv6_bind_works():
    cfile = create_config(('listen=:::9000',))
    Config(cfile).load()
    with mock.patch('socket.socket.bind') as mock_sock:
        server('::', 9000, socket.AF_INET6, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_ipv4_tls_bind_fails():
    cfile = create_config(('tls_listen=127.0.0.1:9000',))
    Config(cfile).load()
    with mock.patch('socket.socket.bind',
                    side_effect=OSError) as mock_sock, \
            pytest.raises(BlackholeRuntimeException):
        server('127.0.0.1', 9000, socket.AF_INET, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_ipv6_tls_bind_fails():
    cfile = create_config(('tls_listen=:::9000',))
    Config(cfile).load()
    with mock.patch('socket.socket.bind',
                    side_effect=OSError) as mock_sock, \
            pytest.raises(BlackholeRuntimeException):
        server('::', 9000, socket.AF_INET6, {})
    assert mock_sock.called is True
    assert mock_sock.call_count is 1


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_tls_ipv4_bind_works():
    cfile = create_config(('listen=127.0.0.1:25',
                           'tls_listen=127.0.0.1:9000',))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind') as mock_sock, \
            mock.patch('ssl.create_default_context') as mock_ssl:
        server('127.0.0.1', 9000, socket.AF_INET, {}, use_tls=True)
    assert mock_sock.called is True
    assert mock_sock.call_count is 1
    assert mock_ssl.called is True
    assert mock_ssl.call_count is 1


@unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_create_server_tls_ipv6_bind_works():
    cfile = create_config(('listen=:::25',
                           'tls_listen=:::9000',))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind') as mock_sock, \
            mock.patch('ssl.create_default_context') as mock_ssl:
        server('::', 9000, socket.AF_INET6, {}, use_tls=True)
    assert mock_sock.called is True
    assert mock_sock.call_count is 1
    assert mock_ssl.called is True
    assert mock_ssl.call_count is 1


class Grp(mock.MagicMock):
    gr_name = 'testgroup'
    gr_gid = 9000


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setgid():
    cfile = create_config(('group=abc',))
    with mock.patch('grp.getgrnam') as mock_getgrnam, \
            mock.patch('os.setgid') as mock_setgid:
        Config(cfile).load()
        setgid()
    assert mock_getgrnam.called is True
    assert mock_setgid.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setgid_same_group():
    cfile = create_config(('', ))
    with mock.patch('os.setgid'):
        Config(cfile).load()
        assert setgid() is None


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@mock.patch('grp.getgrnam', side_effect=KeyError)
def test_setgid_invalid_group(_):
    cfile = create_config(('group=testgroup',))
    with pytest.raises(SystemExit) as err:
        Config(cfile).load()
        setgid()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setgid_no_perms():
    cfile = create_config(('group=testgroup',))
    with mock.patch('grp.getgrnam', side_effect=PermissionError), \
            pytest.raises(SystemExit) as err:
        Config(cfile).load()
        setgid()
    assert str(err.value) == '77'


class Pwd(mock.MagicMock):
    pwd_uid = 9000


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setuid():
    cfile = create_config(('user=abc',))
    with mock.patch('pwd.getpwnam') as mock_getpwnam, \
            mock.patch('os.setuid') as mock_setuid:
        Config(cfile).load()
        setuid()
    assert mock_getpwnam.called is True
    assert mock_setuid.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setuid_same_user():
    cfile = create_config(('', ))
    with mock.patch('os.setuid'):
        Config(cfile).load()
        assert setuid() is None


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setuid_invalid_user():
    cfile = create_config(('user=testuser',))
    with mock.patch('pwd.getpwnam', side_effect=KeyError), \
            pytest.raises(SystemExit) as err:
        Config(cfile).load()
        setuid()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_setuid_no_perms():
    cfile = create_config(('user=testuser',))
    with mock.patch('pwd.getpwnam', side_effect=PermissionError), \
            pytest.raises(SystemExit) as err:
        Config(cfile).load()
        setuid()
    assert str(err.value) == '77'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_set_pid_permissions():
    pidfile = os.path.join(os.getcwd(), 'pid.pid')
    cfile = create_config(('user=testuser', 'group=testgroup',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()
    with mock.patch('pwd.getpwnam', side_effect=KeyError), \
            pytest.raises(SystemExit) as err:
        pid_permissions()
    assert str(err.value) == '64'
    with mock.patch('grp.getgrgid', side_effect=KeyError), \
            pytest.raises(SystemExit) as err:
        pid_permissions()
    assert str(err.value) == '64'
    with mock.patch('os.chown', side_effect=PermissionError), \
            pytest.raises(SystemExit) as err:
        pid_permissions()
    assert str(err.value) == '64'



import getpass
import grp
import logging
import os
import socket
import unittest
from unittest import mock

import pytest

from blackhole.config import Config, config_test, parse_cmd_args, warn_options
from blackhole.exceptions import ConfigException

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_default():
    with mock.patch('getpass.getuser') as mock_getuser, \
            mock.patch('grp.getgrgid') as mock_getgrgid:
        conf = Config()
    assert conf.config_file == '/etc/blackhole.conf'
    assert mock_getuser.called is True
    assert mock_getuser.call_count is 1
    assert mock_getgrgid.called is True
    assert mock_getgrgid.call_count is 1


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_no_access():
    conf = Config()
    conf.config_file = '/fake/file.conf'
    with mock.patch('os.access', return_value=False) as mock_os_access, \
            pytest.raises(ConfigException):
        conf.load()
    assert mock_os_access.called is True
    assert mock_os_access.call_count is 1


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_load():
    cfile = create_config(('#not=thisline', 'listen=10.0.0.1:1025',
                           '''this won't be added''',
                           'mode=bounce   #default accept'))
    conf = Config(cfile).load()
    assert conf.listen == [('10.0.0.1', 1025, socket.AF_INET, {})]
    assert conf.tls_listen == []
    assert getattr(conf, 'not', None) is None
    assert getattr(conf, 'this', None) is None
    assert getattr(conf, 'default', None) is None


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_load_none():
    conf = Config(None).load()
    assert conf.mode == 'accept'
    assert conf.workers is 1


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_warnings():
    settings = (('tls_listen', (1, 2)), ('tls_dhparams', None),
                ('user', 'root'), ('group', 'root'))
    conf = Args(settings)
    args = Args()
    args.less_secure = True
    conf.args = args
    mmock = mock.MagicMock(spec=logging)
    with mock.patch('os.getuid', return_value=0), \
        mock.patch('os.getgid', return_value=0), \
            mock.patch('logging.getLogger', return_value=mmock):
        warn_options(conf)
    assert mmock.warning.call_count == 3


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestCmdParser(unittest.TestCase):

    def test_default_conf(self):
        parser = parse_cmd_args(['-c/fake/file.conf'])
        assert parser.config_file == '/fake/file.conf'
        parser = parse_cmd_args(['--conf=/fake/file.conf'])
        assert parser.config_file == '/fake/file.conf'

    def test_version(self):
        with pytest.raises(SystemExit) as exc:
            parse_cmd_args(['-v'])
        assert str(exc.value) == '0'
        with pytest.raises(SystemExit) as exc:
            parse_cmd_args(['--version'])
        assert str(exc.value) == '0'

    def test_test(self):
        parser = parse_cmd_args(['-t'])
        assert parser.test is True
        parser = parse_cmd_args(['--test'])
        assert parser.test is True

    def test_debug(self):
        parser = parse_cmd_args(['-d'])
        assert parser.debug is True
        parser = parse_cmd_args(['--debug'])
        assert parser.debug is True

    def test_background(self):
        parser = parse_cmd_args(['-b'])
        assert parser.background is True
        parser = parse_cmd_args(['--background'])
        assert parser.background is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestConfigTest(unittest.TestCase):

    def test_config_test(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        user = getpass.getuser()
        group = grp.getgrgid(os.getgid()).gr_name
        settings = ('listen=0.0.0.0:1205', 'user={}'.format(user),
                    'group={}'.format(group), 'timeout=180',
                    'tls_listen=0.0.0.0:1465',
                    'tls_cert={}'.format(cert), 'tls_key={}'.format(key),
                    'delay=10', 'mode=bounce', 'workers=2')
        cfile = create_config(settings)
        args = Args((('less_secure', True), ('config_file', cfile)))
        with mock.patch('blackhole.config.Config.test_port',
                        return_value=True), \
            mock.patch('blackhole.config.Config.test_tls_port',
                       return_value=True), \
                pytest.raises(SystemExit) as exc:
            config_test(args)
        assert str(exc.value) == '0'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestListen(unittest.TestCase):

    def test_default(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.listen == [('127.0.0.1', 25, socket.AF_INET, {}),
                               ('127.0.0.1', 587, socket.AF_INET, {})]

    def test_localhost(self):
        cfile = create_config(('listen=localhost:25',))
        conf = Config(cfile).load()
        assert conf.listen == [('localhost', 25, socket.AF_INET, {})]

    def test_ipv6_disabled(self):
        cfile = create_config(('listen=:::25',))
        conf = Config(cfile).load()
        conf._listen = [('::', 25, socket.AF_UNSPEC, {})]
        with pytest.raises(ConfigException), \
                mock.patch('socket.has_ipv6', False):
            conf.test_ipv6_support()

    @unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
    def test_ipv6(self):
        cfile = create_config(('listen=:::25',))
        conf = Config(cfile).load()
        assert conf.listen == [('::', 25, socket.AF_INET6, {})]

    def test_no_flags(self):
        cfile = create_config(('listen=:25',))
        conf = Config(cfile).load()
        assert conf.listen == [('', 25, socket.AF_INET, {})]

    def test_mode_flag(self):
        cfile = create_config(('listen=:25 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.listen == [('', 25, socket.AF_INET, {'mode': 'bounce'})]

    def test_delay_flag(self):
        cfile = create_config(('listen=:25 delay=30',))
        conf = Config(cfile).load()
        assert conf.listen == [('', 25, socket.AF_INET, {'delay': '30'})]

    def test_delay_range_flag(self):
        cfile = create_config(('listen=:25 delay=30-50',))
        conf = Config(cfile).load()
        assert conf.listen == [('', 25, socket.AF_INET,
                               {'delay': ('30', '50')})]

    def test_mode_and_delay_range_flag(self):
        cfile = create_config(('listen=:25 delay=15-20 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.listen == [('', 25, socket.AF_INET, {'delay': ('15', '20'),
                                                         'mode': 'bounce'})]

    def test_listen_flags_special_ipv4(self):
        cfile = create_config(('listen=:25 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.flags_from_listener('127.0.0.1', 25) == {'mode': 'bounce'}
        assert conf.flags_from_listener('0.0.0.0', 25) == {'mode': 'bounce'}

    def test_listen_flags_special_ipv6(self):
        cfile = create_config(('listen=:::25 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.flags_from_listener('::1', 25) == {'mode': 'bounce'}


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestPort(unittest.TestCase):

    def test_str_port(self):
        cfile = create_config(('listen=127.0.0.1:abc',))
        with pytest.raises(ConfigException):
            Config(cfile).load()

    def test_lower_than_min(self):
        cfile = create_config(('listen=127.0.0.1:0',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_port()

    def test_larger_than_max(self):
        cfile = create_config(('listen=127.0.0.1:99999',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_port()

    def test_port_under_1024_no_perms(self):
        cfile = create_config(('listen=127.0.0.1:1023',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
                pytest.raises(ConfigException):
            conf.test_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    def test_port_under_1024_with_perms_available(self):
        cfile = create_config(('listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=0) as mock_getuid, \
                mock.patch('socket.socket.bind', return_value=True):
            conf.test_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    @unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
    def test_ipv4_and_ipv6_same_port(self):
        cfile = create_config(('listen=127.0.0.1:9000,:::9000', ))
        conf = Config(cfile).load()
        assert conf.listen == [('127.0.0.1', 9000, socket.AF_INET, {}),
                               ('::', 9000, socket.AF_INET6, {})]

    @unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
    def test_ipv4_and_ipv6_diff_port(self):
        cfile = create_config(('listen=127.0.0.1:9000,:::9001', ))
        conf = Config(cfile).load()
        assert conf.listen == [('127.0.0.1', 9000, socket.AF_INET, {}),
                               ('::', 9001, socket.AF_INET6, {})]

    def test_port_under_1024_with_perms_unavailable(self):
        cfile = create_config(('listen=127.0.0.1:1023',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=0) as mock_getuid, \
            mock.patch('socket.socket.bind',
                       side_effect=OSError(1, 'none')) as mock_socket, \
                pytest.raises(ConfigException):
            conf.test_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1
        assert mock_socket.called is True
        assert mock_socket.call_count is 1

    def test_port_over_1023_available(self):
        cfile = create_config(('listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
                mock.patch('socket.socket.bind', return_value=True):
            conf.test_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    def test_port_over_1023_unavailable(self):
        cfile = create_config(('listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
            mock.patch('socket.socket.bind',
                       side_effect=OSError(1, 'none')) as mock_socket, \
                pytest.raises(ConfigException):
            conf.test_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1
        assert mock_socket.called is True
        assert mock_socket.call_count is 1


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestUser(unittest.TestCase):

    def test_invalid_user(self):
        cfile = create_config(('user=xcbsfbsrwgrwgsgrsgsdgrwty4y4fsg',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_user()

    def test_valid_user(self):
        cfile = create_config(('user={}'.format(getpass.getuser()),))
        conf = Config(cfile).load()
        assert conf.user == getpass.getuser()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestGroup(unittest.TestCase):

    def test_invalid_group(self):
        cfile = create_config(('group=xcbsfbsrwgrwgsgrsgsdgrwty4y4fsg',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_group()

    def test_valid_group(self):
        gname = grp.getgrgid(os.getgid()).gr_name
        cfile = create_config(('group={}'.format(gname)))
        conf = Config(cfile).load()
        assert conf.group == grp.getgrgid(os.getgid()).gr_name


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestTimeout(unittest.TestCase):

    def test_default_timeout(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.timeout == 60

    def test_str_timeout(self):
        cfile = create_config(('timeout=xcbsfbsrwgrwgsgrsgsdgrwty4y4fsg',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_timeout()

    def test_valid_timeout(self):
        cfile = create_config(('timeout=10',))
        conf = Config(cfile).load()
        assert conf.timeout == 10

    def test_timeout_over_180(self):
        cfile = create_config(('timeout=300',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_timeout()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestTlsPort(unittest.TestCase):

    def test_default_tls_port(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.tls_listen == []

    def test_str_tls_port(self):
        cfile = create_config(('tls_listen=127.0.0.1:abc',))
        with pytest.raises(ConfigException):
            Config(cfile).load()

    def test_same_port_tls_port(self):
        cfile = create_config(('listen=127.0.0.1:25',
                               'tls_listen=127.0.0.1:25',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_port()

    def test_valid_tls_port(self):
        cfile = create_config(('tls_listen=127.0.0.1:19',))
        conf = Config(cfile).load()
        assert conf.tls_listen == [('127.0.0.1', 19, socket.AF_INET, {})]

    def test_tls_lower_than_min(self):
        cfile = create_config(('tls_listen=127.0.0.1:0',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_port()

    def test_tls_larger_than_max(self):
        cfile = create_config(('tls_listen=127.0.0.1:99999',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_port()

    def test_tls_under_1024_no_perms(self):
        cfile = create_config(('tls_listen=127.0.0.1:1023',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
                pytest.raises(ConfigException):
            conf.test_tls_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    def test_tls_under_1024_with_perms_available(self):
        cfile = create_config(('tls_listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=0) as mock_getuid, \
                mock.patch('socket.socket.bind', return_value=True):
            conf.test_tls_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    def test_tls_under_1024_with_perms_unavailable(self):
        cfile = create_config(('tls_listen=127.0.0.1:1023',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=0) as mock_getuid, \
            mock.patch('socket.socket.bind',
                       side_effect=OSError(1, 'none')) as mock_socket, \
                pytest.raises(ConfigException):
            conf.test_tls_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1
        assert mock_socket.called is True
        assert mock_socket.call_count is 1

    def test_tls_over_1023_available(self):
        cfile = create_config(('tls_listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
                mock.patch('socket.socket.bind', return_value=True):
            conf.test_tls_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1

    def test_tls_over_1023_unavailable(self):
        cfile = create_config(('tls_listen=127.0.0.1:1024',))
        conf = Config(cfile).load()
        with mock.patch('os.getuid', return_value=9000) as mock_getuid, \
            mock.patch('socket.socket.bind',
                       side_effect=OSError(1, 'none')) as mock_socket, \
                pytest.raises(ConfigException):
            conf.test_tls_port()
        assert mock_getuid.called is True
        assert mock_getuid.call_count is 1
        assert mock_socket.called is True
        assert mock_socket.call_count is 1

    @unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
    def test_ipv4_and_ipv6_same_port(self):
        cfile = create_config(('tls_listen=127.0.0.1:9000,:::9000', ))
        conf = Config(cfile).load()
        assert conf.tls_listen == [('127.0.0.1', 9000, socket.AF_INET, {}),
                                   ('::', 9000, socket.AF_INET6, {})]

    @unittest.skipIf(socket.has_ipv6 is False, 'No IPv6 support')
    def test_ipv4_and_ipv6_diff_port(self):
        cfile = create_config(('tls_listen=127.0.0.1:9000,:::9001', ))
        conf = Config(cfile).load()
        assert conf.tls_listen == [('127.0.0.1', 9000, socket.AF_INET, {}),
                                   ('::', 9001, socket.AF_INET6, {})]


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestTls(unittest.TestCase):

    def test_disabled(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.test_tls_settings() is None

    def test_ipv6_disabled(self):
        cfile = create_config(('tls_listen=:::465',))
        conf = Config(cfile).load()
        conf._tls_listen = [('::', 465, socket.AF_UNSPEC, {})]
        with pytest.raises(ConfigException), \
                mock.patch('socket.has_ipv6', False):
            conf.test_tls_ipv6_support()

    def test_port_no_certkey(self):
        settings = ('tls_listen=127.0.0.1:123',)
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == [('127.0.0.1', 123, socket.AF_INET, {})]
        assert conf.tls_cert is None
        assert conf.tls_key is None

    def test_cert_no_port_key(self):
        cert = create_file('crt.crt')
        settings = ('tls_cert={}'.format(cert),)
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == []
        assert conf.tls_cert == cert
        assert conf.tls_key is None

    def test_key_no_port_cert(self):
        key = create_file('key.key')
        settings = ('tls_key={}'.format(key),)
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == []
        assert conf.tls_cert is None
        assert conf.tls_key == key

    def test_cert_key_no_port(self):
        cert = create_file('crt.crt')
        key = create_file('key.key')
        settings = ('tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == []
        assert conf.tls_cert == cert
        assert conf.tls_key == key

    def test_port_cert_no_key(self):
        cert = create_file('crt.crt')
        settings = ('tls_listen=127.0.0.1:123', 'tls_cert={}'.format(cert),)
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == [('127.0.0.1', 123, socket.AF_INET, {})]
        assert conf.tls_cert == cert
        assert conf.tls_key is None

    def test_port_key_no_cert(self):
        key = create_file('key.key')
        settings = ('tls_listen=127.0.0.1:123', 'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_settings()
        assert conf.tls_listen == [('127.0.0.1', 123, socket.AF_INET, {})]
        assert conf.tls_cert is None
        assert conf.tls_key == key

    def test_port_cert_key(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=127.0.0.1:123', 'tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        conf.test_tls_settings()
        assert conf.tls_listen == [('127.0.0.1', 123, socket.AF_INET, {})]
        assert conf.tls_cert == cert
        assert conf.tls_key == key

    def test_default_dhparam(self):
        cfile = create_config(('', ))
        conf = Config(cfile).load()
        assert conf.tls_dhparams is None

    def test_dhparam_works(self):
        dhparams = create_file('dhparams.pem')
        cfile = create_config(('tls_dhparams={}'.format(dhparams), ))
        conf = Config(cfile).load()
        assert conf.tls_dhparams == dhparams

    def test_dhparam_no_exist(self):
        cfile = create_config(('tls_dhparams=/fake/path/dhparams.pem', ))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_tls_dhparams()

    def test_no_flags(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=:123', 'tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        assert conf.tls_listen == [('', 123, socket.AF_INET, {})]

    def test_mode_flag(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=:123 mode=bounce', 'tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        assert conf.tls_listen == [('', 123, socket.AF_INET,
                                   {'mode': 'bounce'})]

    def test_delay_flag(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=:123 delay=30', 'tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        assert conf.tls_listen == [('', 123, socket.AF_INET, {'delay': '30'})]

    def test_delay_range_flag(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=:123 delay=30-50', 'tls_cert={}'.format(cert),
                    'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        assert conf.tls_listen == [('', 123, socket.AF_INET,
                                   {'delay': ('30', '50')})]

    def test_mode_and_delay_range_flag(self):
        key = create_file('key.key')
        cert = create_file('crt.crt')
        settings = ('tls_listen=:123 mode=bounce delay=15-20',
                    'tls_cert={}'.format(cert), 'tls_key={}'.format(key))
        cfile = create_config(settings)
        conf = Config(cfile).load()
        assert conf.tls_listen == [('', 123, socket.AF_INET,
                                   {'delay': ('15', '20'), 'mode': 'bounce'})]

    def test_tls_listen_flags_special_ipv4(self):
        cfile = create_config(('tls_listen=:465 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.flags_from_listener('127.0.0.1', 465) == {'mode': 'bounce'}
        assert conf.flags_from_listener('0.0.0.0', 465) == {'mode': 'bounce'}

    def test_tls_listen_flags_special_ipv6(self):
        cfile = create_config(('listen=:::465 mode=bounce',))
        conf = Config(cfile).load()
        assert conf.flags_from_listener('::1', 465) == {'mode': 'bounce'}


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestDelay(unittest.TestCase):

    def test_no_delay(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.test_delay() is None
        assert conf.delay is None

    def test_delay_longer_than_timeout(self):
        cfile = create_config(('timeout=10', 'delay=20'))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            assert conf.test_delay()
        assert conf.delay > conf.timeout

    def test_delay(self):
        cfile = create_config(('timeout=30', 'delay=5'))
        conf = Config(cfile).load()
        assert conf.timeout > conf.delay

    def test_delay_over_60(self):
        cfile = create_config(('timeout=70', 'delay=70'))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_delay()
        assert conf.delay is 70


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestMode(unittest.TestCase):

    def test_default(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.mode == 'accept'

    def test_invalid_mode(self):
        cfile = create_config(('mode=kura',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            assert conf.test_mode()
        assert conf.mode == 'kura'

    def test_accept(self):
        cfile = create_config(('mode=accept',))
        conf = Config(cfile).load()
        assert conf.mode == 'accept'

    def test_bounce(self):
        cfile = create_config(('mode=bounce',))
        conf = Config(cfile).load()
        assert conf.mode == 'bounce'

    def test_random(self):
        cfile = create_config(('mode=random',))
        conf = Config(cfile).load()
        assert conf.mode == 'random'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestMaxMessageSize(unittest.TestCase):

    def test_no_size(self):
        cfile = create_config(('',))
        conf = Config(cfile).load()
        assert conf.max_message_size == 512000

    def test_invalid_size(self):
        cfile = create_config(('max_message_size=abc',))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_max_message_size()

    def test_size(self):
        cfile = create_config(('max_message_size=1024000',))
        conf = Config(cfile).load()
        assert conf.max_message_size == 1024000
        assert conf.test_max_message_size() is None


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestPidfile(unittest.TestCase):

    def test_pidfile_default(self):
        cfile = create_config(('', ))
        conf = Config(cfile).load()
        assert conf.pidfile == '/tmp/blackhole.pid'

    def test_pidfile_no_permission(self):
        cfile = create_config(('pidfile=/fake/path.pid', ))
        conf = Config(cfile).load()
        with pytest.raises(ConfigException):
            conf.test_pidfile()

    def test_pidfile_with_permission(self):
        cfile = create_config(('pidfile=/tmp/path.pid', ))
        conf = Config(cfile).load()
        with mock.patch('builtins.open', return_value=True):
            conf.test_pidfile()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestDynamicSwitch(unittest.TestCase):

    def test_dynamic_switch_default(self):
        cfile = create_config(('', ))
        conf = Config(cfile).load()
        assert conf.dynamic_switch is True

    def test_dynamic_switch_false(self):
        cfile = create_config(('dynamic_switch=false', ))
        conf = Config(cfile).load()
        assert conf.dynamic_switch is False

    def test_dynamic_switch_invalid(self):
        cfile = create_config(('dynamic_switch=abc', ))
        with pytest.raises(ConfigException):
            conf = Config(cfile).load()
        cfile = create_config(('', ))
        conf = Config(cfile).load()
        conf._dynamic_switch = 'abc'
        with pytest.raises(ConfigException):
            conf.test_dynamic_switch()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
class TestWorkers(unittest.TestCase):

    def test_default(self):
        conf = Config(None).load()
        assert conf.workers == 1

    def test_more_than_cpus(self):
        conf = Config(None).load()
        with mock.patch('multiprocessing.cpu_count', return_value=0), \
                pytest.raises(ConfigException):
            conf.test_workers()

    def test_ok(self):
        cfile = create_config(('workers=4', ))
        conf = Config(cfile).load()
        with mock.patch('multiprocessing.cpu_count', return_value=4):
            conf.test_workers()
        assert conf.workers is 4



import logging

import pytest

from blackhole.logs import configure_logs

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_default():
    args = Args((('debug', False), ('test', False), ('quiet', False)))
    logger = logging.getLogger('blackhole')
    configure_logs(args)
    assert logger.handlers[0].level is logging.INFO


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_debug():
    args = Args((('debug', True), ('test', False), ('quiet', False)))
    logger = logging.getLogger('blackhole')
    configure_logs(args)
    assert logger.handlers[0].level is logging.DEBUG


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_test():
    args = Args((('debug', False), ('test', True), ('quiet', False)))
    logger = logging.getLogger('blackhole')
    configure_logs(args)
    assert logger.handlers[0].level is logging.INFO


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_quiet():
    args = Args((('debug', False), ('test', False), ('quiet', True)))
    logger = logging.getLogger('blackhole')
    configure_logs(args)
    assert logger.handlers[0].level is logging.ERROR



from io import StringIO
from unittest import mock

import pytest

from blackhole.utils import mailname, message_id, get_version

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_mail_name_file():
    check_value = 'file.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('builtins.open', return_value=StringIO(check_value)):
        mn = mailname()
        assert mn == check_value


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_mail_name_socket():
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=False), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname()
    assert mn == check_value


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_mail_name_file_length_0():
    mnfile = create_file('mailname')
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname(mnfile)
    assert mn == check_value


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_mail_name_file_garbage():
    mnfile = create_file('mailname', '            \n     ')
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname(mnfile)
    assert mn == check_value


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_message_id():
    with mock.patch('time.time',
                    return_value=1463290829.4173775) as mock_time, \
        mock.patch('os.getpid', return_value=9000) as mock_getpid, \
            mock.patch('random.getrandbits',
                       return_value=17264867586200823825) as mock_randbits:
        ex_mid = '<{}.{}.{}@blackhole.io>'.format(int(1463290829.4173775 *
                                                  100),
                                                  9000, 17264867586200823825)
        assert message_id('blackhole.io') == ex_mid
    assert mock_time.called is True
    assert mock_getpid.called is True
    assert mock_randbits.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_get_version():
    version_file = create_file('version.py', '__version__ = "9.9.9"')
    with mock.patch('os.path.join', return_value=version_file):
        assert get_version() == '9.9.9'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_get_version_no_access():
    with mock.patch('os.access', return_value=False), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'No __init__.py file found'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_get_version_bad_version_number():
    version_file = create_file('version.py', '__version__ = "abc"')
    with mock.patch('os.path.join', return_value=version_file), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'No valid __version__ string found'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_get_version_version_not_found():
    version_file = create_file('version.py', 'version = "abc"')
    with mock.patch('os.path.join', return_value=version_file), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'No __version__ assignment found'



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides control functionality, including socket wrappers."""

import grp
import logging
import os
import pwd
import socket
try:
    import ssl
except ImportError:
    ssl = None

from .config import Config
from .exceptions import BlackholeRuntimeException


__all__ = ('pid_permissions', 'server', 'setgid', 'setuid')


logger = logging.getLogger('blackhole.control')
ciphers = ['ECDHE-ECDSA-AES256-GCM-SHA384', 'ECDHE-RSA-AES256-GCM-SHA384',
           'ECDHE-ECDSA-CHACHA20-POLY1305', 'ECDHE-RSA-CHACHA20-POLY1305',
           'ECDHE-ECDSA-AES128-GCM-SHA256', 'ECDHE-RSA-AES128-GCM-SHA256',
           'ECDHE-ECDSA-AES256-SHA384', 'ECDHE-RSA-AES256-SHA384',
           'ECDHE-ECDSA-AES128-SHA256', 'ECDHE-RSA-AES128-SHA256']


def _context(use_tls=False):
    """
    Create a TLS context using the certificate, key and dhparams file.

    :param use_tls: Whether to create a TLS context or not.
    :type use_tls: :any:`bool`
    :returns: A TLS context or none.
    :rtype: :any:`ssl.SSLContext` or :any:`None`.

    .. note::

       Created with:

       - :any:`ssl.OP_NO_SSLv2`
       - :any:`ssl.OP_NO_SSLv3`
       - :any:`ssl.OP_NO_COMPRESSION`
       - :any:`ssl.OP_CIPHER_SERVER_PREFERENCE`

       Also responsible for loading Diffie Hellman ephemeral parameters if
       they're provided -- :any:`ssl.SSLContext.load_dh_params`

       If the ``-ls`` or ``--less-secure`` option is provided,
       :any:`ssl.OP_SINGLE_DH_USE` and :any:`ssl.OP_SINGLE_ECDH_USE` will be
       omitted from the context. --
       https://blackhole.io/command-line-options.html#command-line-options --
       added in :ref:`2.0.13`
    """
    if use_tls is False:
        return None
    config = Config()
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ctx.load_cert_chain(config.tls_cert, config.tls_key)
    ctx.options |= ssl.OP_NO_SSLv2
    ctx.options |= ssl.OP_NO_SSLv3
    ctx.options |= ssl.OP_NO_COMPRESSION
    ctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
    if not config.args.less_secure:
        ctx.options |= ssl.OP_SINGLE_DH_USE
        ctx.options |= ssl.OP_SINGLE_ECDH_USE
    ctx.set_ciphers(':'.join(ciphers))
    if config.tls_dhparams:
        ctx.load_dh_params(config.tls_dhparams)
    return ctx


def _socket(addr, port, family):
    """
    Create a socket, bind and listen.

    :param addr: The address to use.
    :type addr: :any:`str`
    :param port: The port to use.
    :type port: :any:`int`
    :param family: The type of socket to use.
    :type family: :any:`socket.AF_INET` or :any:`socket.AF_INET6`.
    :returns: A bound socket.
    :rtype: :any:`socket.socket`
    :raises: :any:`blackhole.exceptions.BlackholeRuntimeException`
    """
    sock = socket.socket(family, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    except (AttributeError, OSError):
        pass
    if family == socket.AF_INET6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
    try:
        sock.bind((addr, port))
    except OSError:
        msg = 'Cannot bind to {}:{}.'.format(addr, port)
        logger.fatal(msg)
        sock.close()
        raise BlackholeRuntimeException(msg)
    os.set_inheritable(sock.fileno(), True)
    sock.listen(1024)
    sock.setblocking(False)
    return sock


def server(addr, port, family, flags={}, use_tls=False):
    """
    A socket and possibly a TLS context.

    Create an instance of :any:`socket.socket`, bind it and return a dictionary
    containing the socket object and a TLS context if configured.

    :param addr: The address to use.
    :type addr: :any:`str`
    :param port: The port to use.
    :type port: :any:`int`
    :param family: The type of socket to use.
    :type family: :any:`socket.AF_INET` or :any:`socket.AF_INET6`.
    :param flags: Flags to use.
    :type flags: :any:`dict`. Default: {}
    :param use_tls: Whether to create a TLS context or not.
    :type use_tls: :any:`bool`
    :returns: A bound socket, a TLS context if configured and any configured
              flags.
    :rtype: :any:`dict`
    """
    sock = _socket(addr, port, family)
    ctx = _context(use_tls=use_tls)
    return {'sock': sock, 'ssl': ctx, 'flags': flags}


def pid_permissions():
    """
    Change the pid file ownership.

    Called before :any:`blackhole.control.setgid` and
    :any:`blackhole.control.setuid` are called to stop
    :any:`blackhole.daemon.Daemon` losing permissions to modify the pid file.

    :raises: :any:`SystemExit` -- :any:`os.EX_USAGE`
    """
    config = Config()
    try:
        user = pwd.getpwnam(config.user)
        group = grp.getgrnam(config.group)
        os.chown(config.pidfile, user.pw_uid, group.gr_gid)
    except (KeyError, PermissionError):
        logger.error('Unable to change pidfile ownership permissions.')
        raise SystemExit(os.EX_USAGE)


def setgid():
    """
    Change group.

    Change to a less privileged group. Unless you're using it incorrectly --
    in which case, don't use it.
    :raises: :any:`SystemExit` -- :any:`os.EX_USAGE` or :any:`os.EX_NOPERM`

    .. note::

       MUST be called BEFORE setuid, not after.
    """
    config = Config()
    try:
        gid = grp.getgrnam(config.group).gr_gid
        os.setgid(gid)
    except KeyError:
        logger.error('Group \'%s\' does not exist.', config.group)
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error('You do not have permission to switch to group \'%s\'.',
                     config.group)
        raise SystemExit(os.EX_NOPERM)


def setuid():
    """
    Change user.

    Change to a less privileged user.Unless you're using it incorrectly --
    inwhich case, don't use it.
    :raises: :any:`SystemExit` -- :any:`os.EX_USAGE` or :any:`os.EX_NOPERM`

    .. note::

       MUST be called AFTER setgid, not before.
    """
    config = Config()
    try:
        uid = pwd.getpwnam(config.user).pw_uid
        os.setuid(uid)
    except KeyError:
        logger.error('User \'%s\' does not exist.', config.user)
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error('You do not have permission to switch to user \'%s\'.',
                     config.user)
        raise SystemExit(os.EX_NOPERM)



import os
from unittest import mock

import pytest

from blackhole.application import run
from blackhole.config import Config
from blackhole.exceptions import (ConfigException, DaemonException,
                                  BlackholeRuntimeException)

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_test():
    cfile = create_config(('', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
        mock.patch('blackhole.config.Config.test_port', return_value=True), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_test_fails():
    cfile = create_config(('listen=127.0.0.1:0', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_load_test_fails():
    cfile = create_config(('listen=127.0.0.1:0', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
        mock.patch('blackhole.config.Config.test',
                   side_effect=ConfigException()), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_foreground():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('blackhole.control.pid_permissions'), \
        mock.patch('blackhole.control.setgid'), \
        mock.patch('blackhole.control.setuid'), \
        mock.patch('blackhole.supervisor.Supervisor.run'), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_foreground_pid_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('os.getpid', return_value=1234), \
        mock.patch('atexit.register', side_effect=DaemonException), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_foreground_socket_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('blackhole.supervisor.Supervisor.close_socks'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers',
                   side_effect=BlackholeRuntimeException), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '77'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_background():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile), '-b']), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('blackhole.daemon.Daemon.daemonize'), \
        mock.patch('blackhole.control.pid_permissions'), \
        mock.patch('blackhole.control.setgid'), \
        mock.patch('blackhole.control.setuid'), \
        mock.patch('blackhole.supervisor.Supervisor.run'), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_run_daemon_daemonize_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile), '-b']), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('os.fork', side_effect=OSError), \
        mock.patch('blackhole.supervisor.Supervisor.'
                   'close_socks') as mock_close, \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '77'
    assert mock_close.called is True



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utility functionality."""

import os
import random
import re
import socket
import time


__all__ = ('mailname', 'message_id', 'get_version')


def mailname(mailname_file='/etc/mailname'):
    """
    A fully qualified domain name for HELO and EHLO.

    :param mailname_file: A path to the mailname file.
    :type mailname_file: :any:`str`
    :returns: A domain name.
    :rtype: :any:`str`

    .. note::

       Prefers content of `mailname_file`, falls back on :any:`socket.getfqdn`
       if `mailname_file` does not exist or cannot be opened for reading.
    """
    if os.access(mailname_file, os.R_OK):
        mailname_content = open(mailname_file, 'r').readlines()
        if len(mailname_content) == 0:
            return socket.getfqdn()
        mailname_content = mailname_content[0].strip()
        if mailname_content != '':
            return mailname_content
    return socket.getfqdn()


def message_id(domain):
    """
    Return a string suitable for RFC 2822 compliant Message-ID.

    :param domain: A fully qualified domain.
    :type domain :any:`str`
    :returns: An RFC 2822 compliant Message-ID.
    :rtype: :any:`str`
    """
    timeval = int(time.time() * 100)
    pid = os.getpid()
    randint = random.getrandbits(64)
    return '<{}.{}.{}@{}>'.format(timeval, pid, randint, domain)


def get_version():
    """
    Extract the __version__ from a file without importing it.

    :return: The version that was extracted.
    :rtype: :any:`str`
    :raises: :any:`AssertionError`
    """
    path = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(path, '__init__.py')
    pattern = re.compile(r'(?P<version>\d+\.\d+(?:\.\d+)?(?:(?:a|b|rc)\d+)?)')
    if not os.access(filepath, os.R_OK):
        raise AssertionError('No __init__.py file found')
    with open(filepath) as fp:
        for line in fp:
            if line.startswith('__version__'):
                mo = pattern.search(line)
                assert mo, 'No valid __version__ string found'
                return mo.group('version')
    raise AssertionError('No __version__ assignment found')



# (The MIT License)
#
# Copyright (c) 2016 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality to spawn and control child processes."""


import asyncio
import logging
import os
import signal

from . import protocols
from .smtp import Smtp
from .streams import StreamProtocol


__all__ = ('Child', )


logger = logging.getLogger('blackhole.child')


class Child:
    """
    A child process.

    Each child process maintains a list of the internal
    :any:`asyncio.create_server` instances it utilises. Each child also
    maintains a list of all connections being managed by the child.
    """

    _started = False
    servers = []
    """List of :any:`asyncio.create_server` instances."""

    clients = []
    """List of clients connected to this process."""

    def __init__(self, up_read, down_write, socks, idx):
        """
        Initialise a child process.

        :param up_read: A pipe to the worker process for heartbeats.
        :type up_read: :any:`os.pipe`
        :param down_write: A pipe to the worker process for heartbeats.
        :type down_write: :any:`os.pipe`
        :param socks: A list of sockets to get data from.
        :type socks: :any:`list`
        """
        self.up_read = up_read
        self.down_write = down_write
        self.socks = socks
        self.idx = idx

    def start(self):
        """Start the child process."""
        logger.debug('Starting child %s', self.idx)
        self._started = True
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        signal.signal(signal.SIGTERM, self.stop)
        self.heartbeat_task = asyncio.Task(self.heartbeat())
        self.loop.run_forever()
        self.stop()
        os._exit(os.EX_OK)

    async def _start(self):
        """Create an asyncio 'server' for each socket."""
        for sock in self.socks:
            server = await self.loop.create_server(lambda: Smtp(self.clients),
                                                   **sock)
            self.servers.append(server)

    def stop(self, signum=None, frame=None):
        """
        Stop the child process.

        Mark the process as being stopped, closes each client connected via
        this child, cancels internal communication with the supervisor and
        finally stops the process and exits.

        :param signum: A signal number.
        :type signum: :any:`int`
        :param frame: Interrupted stack frame.
        :type frame: :any:`frame`
        :returns: Exit the process.
        :rtype: :any:`os._exit` -- :any:`os.EX_OK`
        """
        self.loop.stop()
        for _ in range(len(self.clients)):
            client = self.clients.pop()
            client.close()
        for _ in range(len(self.servers)):
            server = self.servers.pop()
            server.close()
        self.heartbeat_task.cancel()
        self.server_task.cancel()
        for task in asyncio.Task.all_tasks(self.loop):
            task.cancel()
        self._started = False
        os._exit(os.EX_OK)

    async def heartbeat(self):
        """
        Handle heartbeat between a worker and child.

        If a child process stops communicating with it's worker, it will be
        killed, the worker managing it will also be removed and a new worker
        and child will be spawned.

        .. note::

           3 bytes are used in the communication channel.

           - b'x01' -- :any:`blackhole.protocols.PING`
           - b'x02' -- :any:`blackhole.protocols.PONG`

           These message values are defined in the :any:`blackhole.proto`
           schema. Documentation is available at --
           https://blackhole.io/api-protocols.html#blackhole.proto
        """
        read_fd = os.fdopen(self.up_read, 'rb')
        r_trans, r_proto = await self.loop.connect_read_pipe(StreamProtocol,
                                                             read_fd)
        write_fd = os.fdopen(self.down_write, 'wb')
        w_trans, w_proto = await self.loop.connect_write_pipe(StreamProtocol,
                                                              write_fd)
        reader = r_proto.reader
        writer = asyncio.StreamWriter(w_trans, w_proto, reader, self.loop)
        self.server_task = asyncio.Task(self._start())

        while self._started:
            try:
                msg = await reader.read(3)
            except:
                break
            if msg == protocols.PING:
                logger.debug('child.%s.heartbeat: Ping request received from '
                             'parent', self.idx)
                writer.write(protocols.PONG)
        r_trans.close()
        w_trans.close()
        self.stop()



import asyncio
import socket

import pytest

from blackhole.control import server
from blackhole.worker import Worker

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_ping_pong():
    aserver = server('127.0.0.1', 0, socket.AF_INET)
    worker = Worker('1', [aserver, ])
    await asyncio.sleep(35)
    worker.stop()
    assert worker._started is False



import asyncio
import os
from io import BytesIO
import time
from unittest import mock

import pytest

from blackhole import protocols
from blackhole.config import Config
from blackhole.worker import Worker

from ._utils import (cleandir, reset_conf, reset_daemon, reset_supervisor,
                     create_config, create_file, Args)


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_init_start():
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        Worker([], [])
    assert mock_start.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_stop():
    with mock.patch('blackhole.worker.Worker.start') as mock_start, \
            mock.patch('os.kill') as mock_kill:
        w = Worker([], [])
        w.chat_task = mock.Mock()
        w.heartbeat_task = mock.Mock()
        w.rtransport = mock.Mock()
        w.wtransport = mock.Mock()
        w.pid = 123
        w.stop()
    assert mock_start.called is True
    assert mock_kill.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_stop_runtime_exception():
    with mock.patch('blackhole.worker.Worker.start') as mock_start, \
            mock.patch('os.kill', side_effect=ProcessLookupError):
        w = Worker([], [])
        w.chat_task = mock.Mock()
        w.heartbeat_task = mock.Mock()
        w.rtransport = mock.Mock()
        w.wtransport = mock.Mock()
        w.pid = 123
        w.stop()
    assert mock_start.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_parent_start():
    with mock.patch('os.pipe', return_value=('', '')) as mock_pipe, \
        mock.patch('os.fork', return_value=123) as mock_fork, \
        mock.patch('os.close') as mock_close, \
        mock.patch('asyncio.ensure_future') as mock_async, \
            mock.patch('blackhole.worker.Worker.connect') as mock_connect:
        w = Worker([], [])
        w.pid = 123
    assert mock_pipe.call_count == 2
    assert mock_fork.called is True
    assert mock_close.call_count == 2
    assert mock_async.called is True
    assert mock_connect.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_child_start():
    with mock.patch('os.fork', return_value=-1) as mock_fork, \
        mock.patch('blackhole.control.setgid'), \
        mock.patch('blackhole.control.setuid'), \
        mock.patch('asyncio.Task'), \
        mock.patch('blackhole.child.Child.heartbeat'), \
        mock.patch('asyncio.unix_events._UnixSelectorEventLoop.run_forever'), \
        mock.patch('blackhole.child.Child.stop'), \
            mock.patch('os._exit') as mock_exit:
        Worker([], [])
    assert mock_fork.called is True
    assert mock_exit.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_child_start_setgid_fails_invalid_group():
    cfile = create_config(('user=fgqewgreghrehgerhehw',
                           'group=fgqewgreghrehgerhehw'))
    Config(cfile).load()
    with mock.patch('os.pipe', return_value=('', '')), \
        mock.patch('os.fork', return_value=False), \
        mock.patch('os.close'), \
        mock.patch('os.setgid', side_effect=KeyError), \
            pytest.raises(SystemExit) as err:
        Worker([], [])
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_child_start_setgid_fails_permissions():
    cfile = create_config(('user=fgqewgreghrehgerhehw',
                           'group=fgqewgreghrehgerhehw'))
    Config(cfile).load()
    with mock.patch('os.pipe', return_value=('', '')), \
        mock.patch('os.fork', return_value=False), \
        mock.patch('os.close'), \
        mock.patch('os.setgid', side_effect=PermissionError), \
            pytest.raises(SystemExit) as err:
        Worker([], [])
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_child_start_setuid_fails_invalid_user():
    cfile = create_config(('user=fgqewgreghrehgerhehw',
                           'group=fgqewgreghrehgerhehw'))
    Config(cfile).load()
    with mock.patch('os.pipe', return_value=('', '')), \
        mock.patch('os.fork', return_value=False), \
        mock.patch('os.close'), \
        mock.patch('os.setuid', side_effect=KeyError), \
            pytest.raises(SystemExit) as err:
        Worker([], [])
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
def test_child_start_setuid_fails_permissions():
    cfile = create_config(('user=fgqewgreghrehgerhehw',
                           'group=fgqewgreghrehgerhehw'))
    Config(cfile).load()
    with mock.patch('os.pipe', return_value=('', '')), \
        mock.patch('os.fork', return_value=False), \
        mock.patch('os.close'), \
        mock.patch('os.setuid', side_effect=PermissionError), \
            pytest.raises(SystemExit) as err:
        Worker([], [])
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_heartbeat_not_started(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    worker._started = False
    with mock.patch('blackhole.worker.Worker.start') as mock_heartbeat_start, \
            mock.patch('blackhole.worker.Worker.stop') as mock_heartbeat_stop:
        await worker.heartbeat(None)
    assert mock_heartbeat_start.called is False
    assert mock_heartbeat_stop.called is False


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_heartbeat_started_bad_ping(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    worker._started = True
    worker.ping = time.monotonic() - 120
    with mock.patch('blackhole.worker.Worker.start') as mock_heartbeat_start, \
            mock.patch('blackhole.worker.Worker.stop') as mock_heartbeat_stop:
        await worker.heartbeat(None)
    assert mock_heartbeat_start.called is True
    assert mock_heartbeat_stop.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_heartbeat_started_good_ping(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    worker._started = True
    worker.ping = time.monotonic() + 120
    writer = BytesIO()
    async def reset():
        await asyncio.sleep(20)
        worker._started = False
    reset_task = asyncio.Task(reset())
    await worker.heartbeat(writer)
    assert worker._started is False
    reset_task.cancel()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_chat_not_started(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    worker._started = False
    with mock.patch('blackhole.worker.Worker.start') as mock_chat_start, \
            mock.patch('blackhole.worker.Worker.stop') as mock_chat_stop:
        await worker.chat(None)
    assert mock_chat_start.called is False
    assert mock_chat_stop.called is False


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_chat_started_restart(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    worker._started = True
    with mock.patch('blackhole.worker.Worker.start') as mock_chat_start, \
        mock.patch('blackhole.worker.Worker.stop') as mock_chat_stop, \
            mock.patch('asyncio.StreamReader.read', side_effect=Exception):
        reader = asyncio.StreamReader()
        await worker.chat(reader)
    assert mock_chat_start.called is True
    assert mock_chat_stop.called is True


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_chat_started_good_message(event_loop):
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [], loop=event_loop)
    assert mock_start.called is True
    worker._started = True
    worker.ping = 987654321
    reader = asyncio.StreamReader()
    async def reset():
        reader.feed_data(protocols.PONG)
        worker._started = False
    reset_task = asyncio.Task(reset())
    with mock.patch('time.monotonic', return_value=123456789):
        await worker.chat(reader)
    assert worker._started is False
    assert worker.ping == 123456789
    reset_task.cancel()


@pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
                         'cleandir')
@pytest.mark.asyncio
async def test_worker_connect(event_loop):
    up_read, up_write = os.pipe()
    down_read, down_write = os.pipe()
    os.close(up_read)
    os.close(down_write)
    with mock.patch('blackhole.worker.Worker.start') as mock_start:
        worker = Worker('1', [])
    assert mock_start.called is True
    with mock.patch('asyncio.ensure_future'), \
        mock.patch('blackhole.worker.Worker.chat') as mock_chat, \
            mock.patch('blackhole.worker.Worker.heartbeat') as mock_heartbeat:
        await worker.connect('1', up_write, down_read)
    assert mock_chat.called is True
    assert mock_heartbeat.called is True



import logging
import os
import tempfile

import pytest

from blackhole.config import Singleton as CSingleton
from blackhole.daemon import Singleton as DSingleton
from blackhole.supervisor import Singleton as SSingleton


logging.getLogger('blackhole').addHandler(logging.NullHandler())


@pytest.fixture()
def cleandir():
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)


@pytest.fixture()
def reset_conf():
    CSingleton._instances = {}


@pytest.fixture()
def reset_daemon():
    DSingleton._instances = {}


@pytest.fixture()
def reset_supervisor():
    SSingleton._instances = {}


def create_config(data):
    cwd = os.getcwd()
    path = os.path.join(cwd, 'test.conf')
    with open(path, 'w') as cfile:
        cfile.write('\n'.join(data))
    return path


def create_file(name, data=''):
    cwd = os.getcwd()
    path = os.path.join(cwd, name)
    with open(path, 'w') as ffile:
        ffile.write(str(data))
    return path


class Args(object):
    def __init__(self, args=None):
        if args is not None:
            for arg in args:
                setattr(self, arg[0], arg[1])
